from properly_util_python.table_helper import TableHelper
from botocore.exceptions import ClientError
import traceback
from boto3.dynamodb.conditions import Key
import time
from properly_util_python.dynamo_data_utils import DynamoData

from properly_util_python.schemas import UserSchema
from properly_util_python.dynamo_helper import DynamoHelperBase

from properly_util_python.data_utils import GuidUrlSafe

import json
from properly_util_python.properly_logging import ProperLogger


class UserAuthHelper:

    OPERATIONS_ROLE_ID = "S4jlPBglYkqccBcWody6hw"
    BASIC_ROLE_ID = "jczqf4GgkEKd1bxtTgpJOg"

    def __init__(self, dynamo_helper: DynamoHelperBase):
        self.dynamo_helper = dynamo_helper
        self.table_helper = TableHelper(self.dynamo_helper)
        self.user_table = self.table_helper.get_table(TableHelper.USER_TABLE_NAME)
        self.logger = ProperLogger("AuthDecorator")

    # for overview see: https://amzn.to/3xOZMjD
    # for specific claims see: http://openid.net/specs/openid-connect-core-1_0.html#StandardClaims
    # one of the best summaries of relevant fields: https://medium.com/@darutk/understanding-id-token-5f83f50fa02e
    def get_user_auth_context_from_event(self, event):

        self.logger.info("processing event", {"event": event})
        self.logger.info("event type is: {} ".format(type(event)))

        if event is None:
            self.logger.info("event is none")
            return None

        requestContext = event.get("requestContext")
        if requestContext is None:
            self.logger.warning("NoContext", "request context is none")
            return None

        authorizer = requestContext.get("authorizer")
        if authorizer is None:
            self.logger.warning("NoAuthorizer", "authorizer is none")
            return None

        claims = authorizer.get("claims")
        if claims is None:
            self.logger.warning("NoClaims", "claims is none")
            return None

        email = claims.get("email")
        if (not isinstance(email, str)) or (email == ""):
            self.logger.warning("NoEmail", "email none, invalid type, or empty")
            return None

        sub = claims.get("sub")
        if (not isinstance(sub, str)) or (sub == ""):
            self.logger.warning("NoSub", "sub none, invalid type, or empty")
            return None

        is_email_verified = claims.get("email_verified")
        if is_email_verified is None:
            self.logger.info("email is verified is none")
            # remove early exit, see note below
            # return None

        # default false
        is_email_verified_bool = False
        if isinstance(is_email_verified, bool) and (is_email_verified):
            # specifically boolean type and true
            self.logger.info("email verified found as boolean true")
            is_email_verified_bool = True
        elif is_email_verified == "true":
            # specifically a string containing true
            self.logger.info("email verified found as string 'true'")
            is_email_verified_bool = True

        if not is_email_verified_bool:
            # Previously: don't accept unconfirmed email
            # return None
            # Now: this check is weakened by allowing jwt without email verified explicitly set
            # reasoning, the connected systems only provide verified emails, but, that could change
            # facebook : https://stackoverflow.com/a/11847514/7632432
            # google : https://stackoverflow.com/a/30357942/7632432
            # amazon : can't confirm without email as currently configured:
            #           https://docs.aws.amazon.com/cognito/latest/developerguide/user-pool-settings-email-phone-verification.html
            self.logger.info("strict mode would not pass login due to verified flag")
            pass

        sub = claims.get("sub")
        identities = json.loads(claims.get("identities")) if claims.get("identities") else None
        provider = identities.get("providerName") if identities is not None else "Email"

        user_auth_context = {"auth_id": sub, "auth_email": email, "provider": provider}

        return user_auth_context

    # From: https://docs.google.com/document/d/1j6hsvKSra6HkwxFRBKnfEy_wr8nCxYn0oSO8Of5SqUE/edit
    #
    #  returns a valid properly_user_id or None if there was an unauthorizied condition

    def find_or_update_verified_user(self, user_auth: dict, requested_properly_user_id: str, role: str = None):

        auth_id = user_auth.get("auth_id")
        auth_email = user_auth.get("auth_email")
        actual_properly_user_id = self.get_properly_id_from_user_auth_id(auth_id)

        if actual_properly_user_id:
            self.logger.info(
                "Existing Verified:  properly user id: {}  auth id: {}".format(actual_properly_user_id, auth_id)
            )

            if role is not None:
                self.add_role_for_user(actual_properly_user_id, role)

            return actual_properly_user_id

        # was not found, so some update will be attempted
        id_to_update = requested_properly_user_id

        actual_properly_user_id = self.get_properly_id_from_user_auth_email(auth_email)
        if actual_properly_user_id:
            self.logger.info(
                "Existing found:  properly user id: {}  with email: {}".format(actual_properly_user_id, auth_email)
            )
            # use the matching id
            id_to_update = actual_properly_user_id

        actual_properly_user_id = self.create_or_update_if_unverified(id_to_update, auth_id, auth_email)
        if not actual_properly_user_id:
            self.logger.error("UpdateFailed",
                "Failed to update properly user id: {}  with user auth: {}. ".format(actual_properly_user_id, user_auth)
            )
            # will occur when something is invalid about the target record to update (some other user claimed,
            # or different email etc.)
            return None

        if role is not None:
            self.add_role_for_user(actual_properly_user_id, role)

        return actual_properly_user_id

    def add_role_for_user(self, properly_user_id: str, role: str):
        if role is None:
            raise ValueError("Must provide a role to add")

        user_key = {"id": properly_user_id}

        seconds_epoch_now = int(time.time())

        update_expression = (
            "set updatedAt = :updatedAt, " "#r = list_append(if_not_exists(#r, :empty_list), :role_list)"
        )

        expression_attributes = {
            ":updatedAt": seconds_epoch_now,
            ":empty_list": [],
            ":role_list": [role],
            ":role_str": role,
        }

        expression_attribute_names = {"#r": "roles"}  # for reserved keyword

        condition_expression = "NOT contains (#r, :role_str)"

        try:
            response = self.user_table.update_item(
                Key=user_key,
                UpdateExpression=update_expression,
                ExpressionAttributeNames=expression_attribute_names,
                ExpressionAttributeValues=expression_attributes,
                ConditionExpression=condition_expression,
            )
        except ClientError as e:
            # one possible expected 'error' - role already exists.  ignore and continue
            if e.response["Error"]["Code"] == "ConditionalCheckFailedException":
                self.logger.info(f"Attempt to add existing role {role} - skipping")
                return None
            self.logger.error("exception traceback: {}".format(traceback.format_exc()))

            raise

    def get_properly_id_from_user_auth_id(self, auth_id: str):

        key_condition = Key("authIdCognito").eq(auth_id)

        response = self.user_table.query(IndexName="authIdCognito-index", KeyConditionExpression=key_condition,)
        items = response.get("Items")
        if len(items) > 1:
            self.logger.error("more than one user assoicated with auth id: {}".format(auth_id))
        item = None
        for item_candidate in items[:1]:
            item = item_candidate  # use the first one found

        if item is None:
            return None

        properly_user_id = item.get("id")

        return properly_user_id

    def get_properly_id_from_user_verified_email(self, email: str):

        key_condition = Key("verifiedEmail").eq(email)

        response = self.user_table.query(IndexName="verifiedEmail-index", KeyConditionExpression=key_condition,)
        items = response.get("Items")
        if len(items) > 1:
            self.logger.error("more than one user associated with verified email: {}".format(email))

        item = None
        for item_candidate in items[:1]:
            item = item_candidate  # use the first one found

        if item is None:
            return None

        properly_user_id = item.get("id")

        return properly_user_id

    def get_properly_id_from_user_auth_email(self, auth_email: str):

        key_condition = Key("unverifiedEmail").eq(auth_email)

        response = self.user_table.query(IndexName="unverifiedEmail-index", KeyConditionExpression=key_condition,)
        items = response.get("Items")
        if len(items) > 1:
            self.logger.error("more than one user associated with unverified email: {}".format(auth_email))

        item = None
        for item_candidate in items[:1]:
            item = item_candidate  # use the first one found

        if item is None:
            return None

        properly_user_id = item.get("id")

        return properly_user_id

    def create_or_update_by_properly_id(self, requested_properly_user_id, user_auth):

        return "INVALID_ID_NOT_IMPLEMENTED"

    def get_user_from_properly_user_id(self, properly_user_id: str):

        response = self.user_table.get_item(Key={"id": properly_user_id})
        item = response.get("Item")
        return item

    def create_or_update_if_unverified(self, properly_user_id, auth_id: str, auth_email: str):
        # if we have a new cognito user but are holding on to an old properly_user_id from a previous
        # login with a different account, re-generate the requested properly_user_id so that account
        # creation in dynamo can succeed
        existing_user = self.get_user_from_properly_user_id(properly_user_id)
        if (
            existing_user is not None
            and existing_user.get("authIdCognito") != auth_id
            and existing_user.get("unverifiedEmail") != auth_email
        ):
            self.logger.info(
                ("Attempted to create account with existing properly user id {} ".format(properly_user_id))
            )
            properly_user_id = GuidUrlSafe.generate()
            self.logger.info("Generated new ID for creation: {} ".format(properly_user_id))

        seconds_epoch_now = int(time.time())

        user_key = {"id": properly_user_id}

        expression_attributes = {":updatedAt": seconds_epoch_now, ":email": auth_email, ":authIdCognito": auth_id}

        update_expression = (
            "set updatedAt = :updatedAt, "
            "verifiedEmail = :email,"
            "authIdCognito = :authIdCognito, "
            "unverifiedEmail = :email, "
            "createdAt = if_not_exists(createdAt,:updatedAt)"
        )

        # condition represents the idea that:
        # This account is unverified or verifed to the same id (can't verify already verified accounts)
        # and if the unverfiedEmail either doesn't exist or is the same value (can't verify unverfied
        # accounts with different email)
        #  condition_expression = '( attribute_not_exists(authCognitoId) OR (authCognitoId = :authCognitoId) ) ' \
        #                         ' AND (attribute_not_exists(unverifiedEmail) OR (unverifiedEmail = :email) )'

        condition_expression = (
            "(  authIdCognito = :authIdCognito OR attribute_not_exists(authIdCognito)  ) "
            " AND (unverifiedEmail = :email OR attribute_not_exists(unverifiedEmail) ) "
        )

        try:

            response = self.user_table.update_item(
                Key=user_key,
                UpdateExpression=update_expression,
                ExpressionAttributeValues=expression_attributes,
                ConditionExpression=condition_expression,
                ReturnValues="ALL_NEW",
            )
        except ClientError as e:
            if e.response["Error"]["Code"] == "ConditionalCheckFailedException":
                return None
            self.logger.error("exception traceback: {}".format(traceback.format_exc()))

            raise

        return properly_user_id

    def is_user_in_role_for_offer(self, properly_user_id, role_id, offer_id):
        # todo: this is a way of implementing "operations manager for calgary can see calgary deals"
        # or "assigned appraiser can see offers they are assigned to"
        # this is not implemented yet, just using global role check until this capability is required
        raise NotImplementedError("roles for a resource is not implemented")

    # accepts None or empty user_id and role_id
    def is_user_in_role(self, properly_user_id: str, role_id: str):

        if not isinstance(properly_user_id, str) or (properly_user_id == ""):
            return False  # None or empty users are in no role

        user_table = self.table_helper.get_table(TableHelper.USER_TABLE_NAME)

        response = user_table.get_item(Key={UserSchema.ID: properly_user_id},)  # Consider: ConsistentRead=True

        user_dict_dynamo = response.get("Item")
        if user_dict_dynamo is None:
            return False  # not in a role if the user doesn't exist

        if role_id == UserAuthHelper.BASIC_ROLE_ID:
            return True  # if the user exists they are automatically in the basic role

        user_dict = DynamoData.to_dict(user_dict_dynamo)

        roles_array = user_dict.get(UserSchema.ROLES)

        if roles_array is None:
            return False

        has_role_flag = role_id in roles_array

        return has_role_flag

    def is_user_identity_correct(self, properly_user_id: str, sub: str, provider: str) -> bool:
        if not isinstance(properly_user_id, str) or (properly_user_id == ""):
            return False

        user_table = self.table_helper.get_table(TableHelper.USER_TABLE_NAME)

        response = user_table.get_item(Key={UserSchema.ID: properly_user_id},)

        user_dict_dynamo = response.get("Item")
        if user_dict_dynamo is None:
            return False  # not a match if the user doesn't exist

        user_dict = DynamoData.to_dict(user_dict_dynamo)

        cognito_id_key = "authIdCognitoGoogle" if provider == "Google" else "authIdCognito"
        user_record_cognito_id = user_dict.get(cognito_id_key)

        is_identity_match = user_record_cognito_id == sub

        return is_identity_match



import os
import time
from contextlib import contextmanager
from functools import wraps
from typing import Callable, Dict, List, Union

import boto3


class MetricsClient:
    def __init__(self, namespace: str, environment: str = None, dimensions: List[Dict] = None):
        self.cloudwatch = boto3.client("cloudwatch")
        self.namespace = namespace
        self.environment = environment or os.environ.get('PROPERLY_STAGE', 'local')

        self.timers = {}

        additional_dimensions = dimensions or []
        environment_dimensions = [
            {
                "Name": "Environment",
                "Value": self.environment,
            },
        ]

        self.dimensions = environment_dimensions + additional_dimensions

    def record_metric(self, metric_name: str, value: Union[float, int], metric_unit: str = "None", dimensions: List[Dict] = None):
        """
        Records a metric count in CloudWatch

        :param metric_name: The name of the metric
        :param value: The value of the metric. The default unit for this value is to count something.
        :param metric_unit: (Optional) The unit of the metric. Defaults to "None" for count types.
        :param dimensions: (Optional) Any additional dimensions to add to the metric.
        """
        additional_dimensions = dimensions or []
        merged_dimensions = self.dimensions + additional_dimensions

        self.cloudwatch.put_metric_data(
            MetricData=[{
                "MetricName": metric_name,
                "Dimensions": merged_dimensions,
                "Unit": metric_unit,
                "Value": value,
            }],
            Namespace=self.namespace,
        )

    @contextmanager
    def timer(self, metric_name: str):
        """
        A context manager timer that records the time to execute the yielded code.
        The duration of the executed context code will be sent to CloudWatch on exit.

        :param metric_name: The name of the metric
        """
        start_time = time.time()

        # The hand-off to timed code
        yield

        stop_time = time.time()
        duration = stop_time - start_time

        # CloudWatch expects "Duration" (a CloudWatch enum value) to be in seconds.
        self.record_metric(
            metric_name=metric_name,
            value=int(duration),
            metric_unit="Duration",
        )

    def function_timer(self, func: Callable) -> Callable:
        """
        A function wrapper timer that records the time to execute the wrapped function.

        The duration of the executed context code will be sent to CloudWatch on function execution exit.
        The name of the metric will be the function name with "-timing" appended.

        :param func: The function to be timed.
        :return: The wrapped function.
        """
        @wraps(func)
        def wrapper(*args, **kwargs):
            metric_name = f"{func.__name__}-timing"
            with self.timer(metric_name=metric_name):
                # Call the wrapped timed function.
                return func(*args, **kwargs)

        return wrapper



import os

from invoke import task


@task
def setup_environment(c):
    """
    Prepares the environment variables for running the tests
    """
    os.environ["AWS_ACCESS_KEY_ID"] = "FakeID"
    os.environ["AWS_SECRET_ACCESS_KEY"] = "FakeKey"
    os.environ["AWS_DEFAULT_REGION"] = "us-west-1"
    os.environ["PROPERLY_STAGE"] = "local"


@task
def black(c, fix=False):
    """
    Run python black
    --fix : run the auto-fixer
    """
    c.run(f"black {'--check ' if not fix else ''}.")


@task
def pycodestyle(c):
    """
    Run pycodestyle
    """
    c.run("pycodestyle ./src ./tests")


@task(setup_environment)
def test(c, k=None):
    """
    Run tests
    """
    c.run(
        f"python -m pytest -v --log-cli-level=INFO ./tests/ --color=yes {f'-k {k}' if k else ''}"
    )


@task(black, pycodestyle)
def lint(c):
    """
    Run all lint commands
    """
    print(
        """
        \033[0;32m==========================================================
        \033[0;32mNo Issues! More effective at de-linting than Scotch-Brite!
        \033[0;32m==========================================================\033[0m
    """
    )


@task
def typecheck(c):
    """
    Run python mypy
    """
    c.run("mypy src tests --config-file=mypy.ini")


@task(aliases=['format'])
def prettify(c):
    """
    Run the auto-formatter
    """
    black(c, fix=True)


@task()
def deploy(c, stage=None):
    """
    Run the serverless deploy
    --stage : stage to deploy to
    """
    if not stage:
        print("must supply a stage, one of dev, staging, prod")
        return

    c.run(f"./node_modules/serverless/bin/serverless.js deploy --stage {stage}")



import warnings

from properly_util_python.dynamo_utils import get_dynamo_item
from properly_util_python.dynamo_helper import DynamoHelperBase
from properly_util_python.table_helper import TableHelper


class PropertyRemapping:

    def __init__(self, dynamo_helper: DynamoHelperBase):
        self.dynamo_helper = dynamo_helper
        self.table_helper = TableHelper(dynamo_helper)

        # Initialize lazy loading offers
        self.offer_id = None
        self.offer_item = None

    def _get_merged_property_data(self, offer_id):
        """
        Combines the property from the user provided property with verified info.
        The verified info, if the data exists, will always overwrite the user provided info.

        The data from the user provided property and verified info are generally the same shape.
        So simple dictionary updates on top of each other should allow us to merged the data.

        This is the priority order of data. The higher data source will overwrite lower data sources.
        1. From Property --> Verified information
        3. From Offer --> Additional information
        2. From Property --> Offer location information
        4. From Offer --> User provided information

        :param offer_id: The offer ID
        :return: The merged property
        """
        offer_item = self.get_offer(offer_id)

        property_id = offer_item['propertyId']
        property_item = self._get_dynamo_item_from_table(TableHelper.PROPERTY_TABLE_NAME, property_id)

        verified_property_dict = property_item.get('verifiedInfo', {})
        property_location_dict = property_item.get('location', {})
        user_supplied_property_dict = offer_item.get('userSuppliedProperty', {})
        additional_information_dict = offer_item.get('additionalInformation', {})

        # Start with an empty combined data
        combined_property_dict = {}

        # 4. Merge in the user provided property information
        combined_property_dict.update(user_supplied_property_dict)

        # 3. Merge in the location information from the Property table
        normalized_location_dict = self._normalize_location_data(property_location_dict)
        combined_property_dict.update(normalized_location_dict)

        # 2. Merge in the additional information
        # This is automatically calculated information and additional insights
        combined_property_dict.update(additional_information_dict)

        # 1. Merge in the verified property information
        # This will overwrite any information from user provided property
        combined_property_dict.update(verified_property_dict)

        return combined_property_dict

    def _normalize_location_data(self, location_dict):
        """
        This converts the location information from the Property table
        to use the same keys as the Property model (as used by verified info and user supplied property)
        :param location_dict: The location data dictionary in the Property table format
        :return: The normalized location data dictionary to fit with the Property model
        """
        normalized_location_dict = {}

        # Address is a special case as the address is written twice in the property model
        location_address_key = 'formattedAddress'
        normalized_address_keys = ['address', 'googleFormattedAddress']

        location_address_value = location_dict.get(location_address_key, None)
        if location_address_value:
            spanned_value_array = [location_address_value] * len(normalized_address_keys)
            normalized_location_dict.update(zip(normalized_address_keys, spanned_value_array))

        normalization_key_map = {
            'lat': 'latitude',
            'lng': 'longitude',
            'placeId': 'googlePlaceId'
        }

        # Loop through the left side of the normalization key map
        for location_type_key in normalization_key_map.keys():
            # Check if the raw location dictionary has the key we want to translate
            if location_type_key not in location_dict:
                continue

            # We found a key we want to change
            # Get the key from the normalization map to the Property model key
            # (the right side of normalization key map)
            normalized_type_key = normalization_key_map[location_type_key]
            # Transfer the value over
            normalized_location_dict.update({normalized_type_key: location_dict[location_type_key]})

        return normalized_location_dict

    def _get_property_summary_dictionary(self, property_dictionary):
        """
        Creates a summarized version of the property with a lot fewer keys.
        This creates the summarized shape for pop_extra_keys=True

        :param property_dictionary: The full data format of the property
        :return: The summarized property with fewer keys
        """
        location_keys = [
            'address',
            'googleFormattedAddress',
            'latitude',
            'longitude',
            'googlePlaceId',
        ]

        normalized_location_dict = {}
        for key in location_keys:
            if key not in property_dictionary:
                continue
            normalized_location_dict[key] = property_dictionary[key]

        return normalized_location_dict

    def _get_dynamo_item_from_table(self, table_name, item_id):
        table_name_with_environment = self.table_helper.get_table_name(table_name)
        return get_dynamo_item(table_name_with_environment, item_id, dynamodb_helper=self.dynamo_helper)

    def get_offer(self, offer_id):
        """
        Gets the offer by ID.
        This will get lazy loaded once per ID.
        If subsequent other calls uses the same offer, it will use the lazy loaded instance.
        This saves multiple calls to the database and allows easy access to the offer from multiple places.
        :param offer_id: The ID of the offer.
        :return: The offer dictionary
        """
        if not self.offer_item or self.offer_id != offer_id:
            self.offer_id = offer_id
            self.offer_item = self._get_dynamo_item_from_table(TableHelper.OFFER_TABLE_NAME, offer_id)
        return self.offer_item

    def get_home_from_offer_id(self, offer_id, *args, **kwargs):
        """
        Gets the merged home property from an offer ID.
        Can accept parameters to return a summarized version of the property.
        Can accept parameters to return the offer in the response.

        If you want to summarized form of the property, use `get_merged_property_summary_from_offer_id()`
        If you want to full form of the property, use `get_merged_property_from_offer_id()`
        If you need the offer, use `get_offer()`.  The offer is cached and will not cause another hit to
        the database if you call `get_offer()`.

        :param offer_id: The offer ID
        :param args: Not used.
        :param kwargs: pop_extra_keys - If true, returns the summarized version of the property.
                                        Otherwise returns all property data.
                       send_offer_data - If true, returns the offer dictionary. Otherwise returns only the property.
        :return: The property and optionally the offer ID.
        """
        # Deprecate this call.  Trying to remove the use of unknown arguments.
        warnings.warn("Call to deprecated function get_home_from_offer_id.",
                      category=DeprecationWarning,
                      stacklevel=2)

        is_summaries = kwargs.get('pop_extra_keys')
        is_sending_offer = kwargs.get('send_offer_data')

        if is_summaries:
            property_data = self.get_merged_property_summary_from_offer_id(offer_id)
        else:
            property_data = self.get_merged_property_from_offer_id(offer_id)

        if is_sending_offer:
            return {
                'offer': self.get_offer(offer_id),
                'home': property_data,
            }
        return property_data

    def get_merged_property_from_offer_id(self, offer_id):
        """
        Returns the merged property data.
        This is the priority order of data. The higher data source will overwrite lower data sources.
        1. From Property --> Verified information
        3. From Offer --> Additional information
        2. From Property --> Offer location information
        4. From Offer --> User provided information

        Use this function instead of `get_home_from_offer_id(offer_id)`

        :param offer_id: The offer ID
        :return: The merged property data
        """
        normalized_property = self._get_merged_property_data(offer_id)
        return normalized_property

    def get_merged_property_summary_from_offer_id(self, offer_id):
        """
        Returns the summarized version of the merged property data.
        This is the priority order of data. The higher data source will overwrite lower data sources.
        1. From Property --> Verified information
        3. From Offer --> Additional information
        2. From Property --> Offer location information
        4. From Offer --> User provided information

        The data that is in the summary are:
        * 'address'
        * 'googleFormattedAddress'
        * 'latitude'
        * 'longitude'
        * 'googlePlaceId'

        Use this function instead of `get_home_from_offer_id(offer_id, pop_extra_keys=True)`

        :param offer_id: The offer ID
        :return: The summary of the merged property data
        """
        normalized_property = self.get_merged_property_from_offer_id(offer_id)
        property_summary = self._get_property_summary_dictionary(normalized_property)
        return property_summary



from properly_util_python.dynamo_helper import EnvironmentControlledDynamoHelper


class ViewConfig:
    """
    The ViewConfig is a convenience wrapper class that contains all the information necessary for
    Flask to fully define a REST endpoint.
    """
    def __init__(self, klass: type, name: str, rule: str, methods: list, **kwargs):
        """
        The Flask REST endpoint wrapper class.
        :param klass: The Flask view class.
        :param name: The ID name of the Flask REST endpoint.
        :param rule: The REST endpoint path rule.
        :param methods: The supported ways to access the Flask REST endpoint.
        :param kwargs: Any keyword arguments to pass to the Flask view at creation time.
        """
        self.klass = klass
        self.name = name
        self.rule = rule
        self.methods = methods
        self.kwargs = kwargs


def add_flask_views(app, view_configs: list):
    """
    The helper method that initializes and adds the Flask REST endpoints defined in the
    ViewConfigs to the Flask app instance.
    :param app: The Flask app instance.
    :param view_configs: The list of ViewConfigs to define REST endpoints.
    """
    for view_config in view_configs:
        view = view_config.klass.as_view(
            name=view_config.name,
            app=app,
            dynamo_helper=EnvironmentControlledDynamoHelper(),
            **view_config.kwargs,
        )
        app.add_url_rule(
            rule=view_config.rule,
            view_func=view,
            methods=view_config.methods
        )



import boto3
import logging
import psycopg2
import psycopg2.extras


logger = logging.getLogger()
logger.level = logging.INFO


class RedshiftHelper:

    # TODO: Initialize from the environment to generalize
    USER = "awsuser"
    DATABASE = "analytics"
    CLUSTER = "properly-general-002"
    ENDPOINT = "properly-general-002.c25yretklzxj.us-east-1.redshift.amazonaws.com"
    PORT = 5439

    def __init__(self, user_to_emulate: str):
        self.conn = None
        self.creds = None
        self.user = user_to_emulate or RedshiftHelper.USER

    def _get_temp_creds(self):
        # Inspiration from:
        # https://blog.mitocgroup.com/aws-lambda-to-redshift-connection-using-iam-authentication-and-nat-gateway-b40c6002082b
        try:
            client = boto3.client('redshift')
            creds = client.get_cluster_credentials(
                DbUser=self.user,
                DbName=RedshiftHelper.DATABASE,
                ClusterIdentifier=RedshiftHelper.CLUSTER,
                DurationSeconds=3600)
        except Exception as ex:
            logger.error("Credentials Issue: {}".format(ex))
            raise

        return creds

    def _create_connection(self):
        try:
            self.conn = psycopg2.connect(
                dbname=RedshiftHelper.DATABASE,
                user=self.creds['DbUser'],
                password=self.creds['DbPassword'],
                port=RedshiftHelper.PORT,
                host=RedshiftHelper.ENDPOINT)
        except Exception as ex:
            logger.error("Connection Issue: {}".format(ex))
            raise

    def _initialize_conn(self):
        self.creds = self._get_temp_creds()

        self._create_connection()

    def create_cursor(self):
        if not self.conn:
            self._initialize_conn()

        cursor = self.conn.cursor(cursor_factory=psycopg2.extras.DictCursor)
        return cursor

    def commit(self):
        if not self.conn:
            self._initialize_conn()
        self.conn.commit()

    def close(self):
        if not (self.conn is None):
            self.conn.close()
            self.conn = None



from shapely.geometry import Point


class ShapelyPoint:
    @staticmethod
    def from_latlong(latitude, longitude):
        return Point(longitude, latitude)


class LatLongPoint:
    @staticmethod
    def from_shapely_point(point: Point):
        return {
            "latitude": point.y,
            "longitude": point.x,
        }




import boto3
import os


class DynamoHelperBase:

    def get_dynamo_db(self):
        raise NotImplementedError()


class DynamoHelper(DynamoHelperBase):

    def get_dynamo_db(self):
        dynamodb = boto3.resource('dynamodb' )
        return dynamodb


class DynamoTestHelper(DynamoHelperBase):

    def get_dynamo_db(self):
        dynamo_alternate_endpoint = "http://localhost:8008"
        dynamodb = boto3.resource('dynamodb', endpoint_url=dynamo_alternate_endpoint, region_name="us-west-1")
        return dynamodb


class EnvironmentControlledDynamoHelper(DynamoHelperBase):

    def __init__(self):
        self.dynamo_helper = DynamoHelper()
        self.test_dynamo_helper = DynamoTestHelper()

    def get_dynamo_db(self):
        use_test_env_flag = os.environ.get('PROPERLY_USE_LOCAL_DYNAMO', None)
        if (use_test_env_flag is None):
            return self.dynamo_helper.get_dynamo_db()
        else:
            return self.test_dynamo_helper.get_dynamo_db()





import functools

from flask import request
from flask.wrappers import Response

from properly_util_python.dynamo_helper import EnvironmentControlledDynamoHelper
from properly_util_python.properly_logging import ProperLogger

from properly_util_python.user_auth import UserAuthHelper


def authorized(role, return_keys: list = []):
    def outer(func):
        @functools.wraps(func)
        def inner(*args, **kwargs):
            logger = ProperLogger("AuthDecorator")
            logger.info("Authorizing request", {"request": request})

            user_auth_helper = UserAuthHelper(EnvironmentControlledDynamoHelper())
            event_object = request.environ.get("event")
            user_auth = user_auth_helper.get_user_auth_context_from_event(event_object)

            if user_auth is None:
                logger.error("NoContext", "no context")
                return Response(response="{'error': 'no context'}", status=403, mimetype="application/json")

            sub = user_auth.get("auth_id")
            provider = user_auth.get("provider")
            logger.add_common_tags({"sub": sub, "provider": provider})

            # find the user
            properly_user_id_caller = user_auth_helper.get_properly_id_from_user_verified_email(
                email=user_auth["auth_email"]
            )

            if properly_user_id_caller is None:
                logger.error("UserNotFound", "user not found")
                return Response(response="{'error': 'user not found'}", status=403, mimetype="application/json")
 
            logger.add_common_tags({"user_id": properly_user_id_caller})

            # verify the user record matches the Cognito information received
            is_user_identity_correct = user_auth_helper.is_user_identity_correct(properly_user_id_caller, sub, provider)
            if not is_user_identity_correct:
                logger.error("MismatchedIdentity", "identity received doesn't match user record")
                return Response(
                    response="{'error': 'identity received doesn't match user record'}",
                    status=403,
                    mimetype="application/json",
                )

            # verify the user has the right roles for the operation
            is_in_role_flag = user_auth_helper.is_user_in_role(properly_user_id_caller, role)

            if not is_in_role_flag:
                logger.error("MissingRole", "missing role")
                return Response(response="{'error': 'missing role'}", status=403, mimetype="application/json")

            logger.info("Authorization succeeded")

            if "user_id" in return_keys:
                return func(*args, user_id=properly_user_id_caller, **kwargs)

            return func(*args, **kwargs)

        return inner

    return outer





import time
from properly_util_python.dynamo_helper import DynamoHelperBase
from properly_util_python.table_helper import TableHelper


class DistributedPropertyTuple:
    META_ID = "id"
    META_DESCRIPTION = "description"
    META_SHORT_NAME = "shortName"
    META_CONFIG_VAL = 'configVal'

    TABLE_UPDATED_AT_MILLIS = 'updatedAtMillis'

    def __init__(self, property_id: str, short_name: str, long_description: str, dynamo_helper: DynamoHelperBase):
        self.dynamo_helper = dynamo_helper
        self.table_helper = TableHelper(dynamo_helper=dynamo_helper)
        self.property_id = property_id
        self.short_name = short_name
        self.long_description = long_description

    def set_property(self, property_val:dict):

        meta_table = self.table_helper.get_table(TableHelper.META_TABLE_NAME)
        now_millis_epoch_utc = int(1000*time.time())

        config_object = {
            DistributedPropertyTuple.META_ID: self.property_id,
            DistributedPropertyTuple.META_SHORT_NAME: self.short_name,
            DistributedPropertyTuple.META_DESCRIPTION: self.long_description,
            DistributedPropertyTuple.TABLE_UPDATED_AT_MILLIS: now_millis_epoch_utc,
            DistributedPropertyTuple.META_CONFIG_VAL: property_val
        }

        meta_table.put_item(Item=config_object)

    def get_property(self):
        meta_table = self.table_helper.get_table(TableHelper.META_TABLE_NAME)

        response = meta_table.get_item(
            Key={
                DistributedPropertyTuple.META_ID: self.property_id,
            })
        config = response.get('Item', None)
        if config is None:
            return None

        last_evaluated_key = config.get(DistributedPropertyTuple.META_CONFIG_VAL)

        return last_evaluated_key





import os

from properly_util_python.dynamo_helper import DynamoHelperBase, DynamoTestHelper

class TableHelper:

    TABLE_TEMPLATE = "{0}-{1}"

    OFFER_TABLE_NAME = "Offer-002"

    PROPERTY_TABLE_NAME = "Property-002"

    USER_TABLE_NAME = "User-002"

    META_TABLE_NAME = "Meta-001"

    HISTORICAL_RAW_NAME = "property-historical-clean-003"

    HISTORICAL_CLEAN_NAME = "property-historical-clean-003"

    def __init__(self, dynamo_helper: DynamoHelperBase):
        self.dynamo_helper = dynamo_helper

    def get_table_name(self, name):
        stage = os.environ.get('PROPERLY_STAGE')
        table_name = TableHelper.TABLE_TEMPLATE.format(stage, name)
        return table_name

    def get_table(self, name):
        table_name = self.get_table_name(name)
        helper = self.dynamo_helper
        dynamodb = helper.get_dynamo_db()
        table = dynamodb.Table(table_name)
        return table


    def build_tables_for_test(self):

        if (not isinstance(self.dynamo_helper, DynamoTestHelper)):
            # do not build tables if it is prod. a safety check
            return

        ddb = self.dynamo_helper.get_dynamo_db()

        table_name = self.get_table_name(TableHelper.META_TABLE_NAME)
        ddb.create_table(
            AttributeDefinitions=[
                {
                    'AttributeName': 'id',
                    'AttributeType': 'S'
                },
            ],
            TableName=table_name,
            KeySchema=[
                {
                    'AttributeName': 'id',
                    'KeyType': 'HASH'
                },
            ],
            ProvisionedThroughput={
                'ReadCapacityUnits': 5,
                'WriteCapacityUnits': 5
            },
        )
        table_name = self.get_table_name(TableHelper.OFFER_TABLE_NAME)
        ddb.create_table(
            AttributeDefinitions=[
                {
                    'AttributeName': 'id',
                    'AttributeType': 'S'
                },
                {
                    'AttributeName': 'propertyId',
                    'AttributeType': 'S'
                },
            ],
            TableName=table_name,
            KeySchema=[
                {
                    'AttributeName': 'id',
                    'KeyType': 'HASH'
                },
            ],
             GlobalSecondaryIndexes=[
                TableHelper.index_from_name_and_attribute('propertyId-index', 'propertyId')
            ],
            ProvisionedThroughput={
                'ReadCapacityUnits': 5,
                'WriteCapacityUnits': 5
            },
        )

        table_name = self.get_table_name(TableHelper.PROPERTY_TABLE_NAME)
        ddb.create_table(
            AttributeDefinitions=[
                {
                    'AttributeName': 'id',
                    'AttributeType': 'S'
                },
                {
                    'AttributeName': 'locationPlaceId',
                    'AttributeType': 'S'
                },
            ],
            TableName=table_name,
            KeySchema=[
                {
                    'AttributeName': 'id',
                    'KeyType': 'HASH'
                },
            ],

            GlobalSecondaryIndexes=[
                TableHelper.index_from_name_and_attribute('LocationPlaceIdIndex', 'locationPlaceId')
            ],

            ProvisionedThroughput={
                'ReadCapacityUnits': 5,
                'WriteCapacityUnits': 5
            },
        )

        table_name = self.get_table_name(TableHelper.USER_TABLE_NAME)
        ddb.create_table(
            AttributeDefinitions=[
                {
                    'AttributeName': 'id',
                    'AttributeType': 'S'
                },
                {
                    'AttributeName': 'authIdCognito',
                    'AttributeType': 'S'
                },
                {
                    'AttributeName': 'unverifiedEmail',
                    'AttributeType': 'S'
                },
                {
                    'AttributeName': 'verifiedEmail',
                    'AttributeType': 'S'
                },
            ],
            TableName=table_name,
            KeySchema=[
                {
                    'AttributeName': 'id',
                    'KeyType': 'HASH'
                },
            ],

            GlobalSecondaryIndexes=[
                TableHelper.index_from_name_and_attribute('authIdCognito-index', 'authIdCognito'),
                TableHelper.index_from_name_and_attribute('unverifiedEmail-index', 'unverifiedEmail'),
                TableHelper.index_from_name_and_attribute('verifiedEmail-index', 'verifiedEmail'),
            ],

            ProvisionedThroughput={
                'ReadCapacityUnits': 5,
                'WriteCapacityUnits': 5
            },
        )


    @classmethod
    def index_from_name_and_attribute(cls, name: str, attribute: str):
        return_val = {
                        'IndexName': name,
                        'KeySchema': [
                            {
                                'AttributeName': attribute,
                                'KeyType': 'HASH'
                            },
                        ],
                        'Projection': {
                            'ProjectionType': 'ALL',
                        },
                        'ProvisionedThroughput': {
                            'ReadCapacityUnits': 5,
                            'WriteCapacityUnits': 5
                        }
                    }
        return return_val






import setuptools
import os

major = 0
minor = 21
build_num = os.environ.get("PROPERLY_BUILD_NUMBER", "0-dev")

VERSION_TEMPLATE = "{}.{}.{}"
version_val = VERSION_TEMPLATE.format(major, minor, build_num)


setuptools.setup(
    name="properly_util_python",
    version=version_val,
    author="GoProperly",
    author_email="info@goproperly.com",
    description="Utility and helper functions for common Properly operations in python.",
    long_description="public",
    long_description_content_type="text/markdown",
    url="https://github.com/GoProperly/properly-util-python",
    packages=setuptools.find_packages(exclude=["custom_scripts", "tests"]),
    install_requires=["orjson >3, <3.6",],
    classifiers=("Programming Language :: Python :: 3",),
)



from flask import Flask
# flask is a peer dependency


class LambdaHelper:

    DEFAULT_HEADERS = {
        'Access-Control-Allow-Origin': '*',
        'Access-Control-Allow-Credentials': True
    }

    MIME_TYPE = 'application/json'

    def __init__(self, flask_app: Flask):
        self.flask_app = flask_app

    def build_flask_response(self, status_code: int, json_body: str):

        response = self.flask_app.response_class(
            response=json_body,
            status=status_code,
            mimetype=LambdaHelper.MIME_TYPE,
            headers=LambdaHelper.DEFAULT_HEADERS
        )

        return response





